#!/bin/bash

######################################################################
# Proprietary Information D-Wave Systems Inc.                        #
# Copyright (c) 2017 by D-Wave Systems Inc. All rights reserved.     #
# D-Wave Systems Inc., 3033 Beta Ave., Burnaby, BC, V5G 4M9, Canada. #
######################################################################

# This shell script (which is all awk) expects to read two things from
# standard input: 1) the geometry file for the current workspace 2)
# the output from dw-get-qmi.  It converts the standard text format
# representation of a QMI into the Qubist representation.  The output
# from this script should be suitable for plugging in to the Data
# sheet of Qubist.  Note that the output from this program is not
# scaled and so will likely need to be scaled to be visualized in
# Qubist.  Of course, Qubist takes care of that for us, and so that's
# why we don't do it here.
#
# In a more perfect world, the code that is shared between this script
# and some related awk scripts would be broken out into a single
# location rather than being replicated.  Refactoring, anyone?

awk '

BEGIN {lines = 0}

$2 ~ /^<==$/ && $1 ~ /^Q[0-9]*$/ {
    qnum = 0 + substr($1,2,4);
    weight = $3;
    line[lines] = sprintf("%d %d %g", qnum, qnum, weight);
    lines = 1 + lines;
    next;
}

$2 ~ /^<==$/ && $1 ~ /^C[0-9]*$/ {
    cnum = 0 + substr($1,2,4);
    Qubits_from_Coupler(cnum);
    weight = $3;
    line[lines] = sprintf("%d %d %g", pq1, pq2, weight);
    lines = 1 + lines;
    next;
}

/^L:/ { DW_L = $2 
        qubits = 2*DW_L*DW_M*DW_N;
        couplers = DW_L*DW_L*DW_M*DW_N + DW_L*DW_M*(DW_N-1) + DW_L*(DW_M-1)*DW_N;
}

/^M:/ { DW_M = $2
        qubits = 2*DW_L*DW_M*DW_N;
        couplers = DW_L*DW_L*DW_M*DW_N + DW_L*DW_M*(DW_N-1) + DW_L*(DW_M-1)*DW_N;
}

/^N:/ { DW_N = $2
        qubits = 2*DW_L*DW_M*DW_N;
        couplers = DW_L*DW_L*DW_M*DW_N + DW_L*DW_M*(DW_N-1) + DW_L*(DW_M-1)*DW_N;
}

/^Q/ { gsub(" ", "");
       split($0, token, ":");
       qubit_range = token[1];
       bit_mask = token[2];
       sub("-.*$", "", qubit_range);
       sub("^Q", "", qubit_range);
       qubit_index = 0 + qubit_range;
       bit_index = 1;
       bits = length(bit_mask);

       while (bit_index <= bits)
       {
           qubit[qubit_index] = substr(bit_mask, bit_index, 1);
           ++ qubit_index;
           ++ bit_index;
       }
}

/^C/ { gsub(" ", "");
       split($0, token, ":");
       coupler_range = token[1];
       bit_mask = token[2];
       sub("-.*$", "", coupler_range);
       sub("^C", "", coupler_range);
       coupler_index = 0 + coupler_range;
       bit_index = 1;
       bits = length(bit_mask);

       while (bit_index <= bits)
       {
           coupler[coupler_index] = substr(bit_mask, bit_index, 1);
           ++ coupler_index;
           ++ bit_index;
       }
}

END {
    printf("%d %d\n", qubits, lines);
    for (i=0; i<lines; ++i) printf("%s\n", line[i]);
}

function ChimInd_from_Linear(linear)
{
    k = (linear % DW_L);
    linear -= k;
    linear /= DW_L;

    u = (linear % 2);
    linear -= u;
    linear /= 2;

    j = (linear % DW_N);
    linear -= j;
    linear /= DW_N;

    i = linear;
}

function Linear_from_ChimInd(i,j,u,k,temp)
{
    return 2*DW_N*DW_L*i + 2*DW_L*j + DW_L*u + k;
}

function Coupler_from_Qubits(pqubit1, pqubit2)
{
    ChimInd_from_Linear(pqubit1);
    i1 = i;
    j1 = j;
    u1 = u;
    k1 = k;

    ChimInd_from_Linear(pqubit2);
    i2 = i;
    j2 = j;
    u2 = u;
    k2 = k;

    # Intracell coupler:
    if (i1 == i2 && j1 == j2 && u1 != u2)
    {
        if (u1 == 0) offset = k1 * DW_L + k2;
        if (u2 == 0) offset = k2 * DW_L + k1;
        return i1 * (DW_L*DW_L*DW_N) + j1 * (DW_L*DW_L) + offset;
    }

    # Horizontal Intercell coupler:
    if (i1 == i2 && ((j1 - j2 == 1) || (j1 - j2 == -1)) && u1 == 1 && u2 == 1 && k1 == k2)
    {
        if (j1 < j2) jmin = j1;
        if (j2 < j1) jmin = j2;
        return DW_L*DW_L*DW_M*DW_N + i1 * (DW_L*(DW_N - 1)) + jmin * DW_L + k1;
    }

    # Vertical Intercell coupler:
    if (((i1 - i2 == 1) || (i1 - i2 == -1)) && j1 == j2 && u1 == 0 && u2 == 0 && k1 == k2)
    {
        if (i1 < i2) imin = i1;
        if (i2 < i1) imin = i2;
        return DW_L*DW_L*DW_M*DW_N + DW_L*DW_M*(DW_N - 1) + imin * (DW_L*DW_N) + j1 * DW_L + k1;
    }

    printf "error: qubits %d and %d are not connected by a coupler\n", pqubit1, pqubit2;
    exit 1;
}

function Qubits_from_Coupler(coupler, temp_coupler, offset, ti, tj, tk0, tk1)
{
    temp_coupler = coupler;

    # Intracell coupler:
    if (coupler < DW_L*DW_L*DW_M*DW_N)
    {
        offset = coupler % (DW_L * DW_L);
        coupler -= offset;
        coupler /= (DW_L * DW_L);

        tj = coupler % DW_N;
        coupler -= tj;
        coupler /= DW_N;

        ti = coupler;

        tk1 = offset % DW_L;
        offset -= tk1;
        tk0 = offset / DW_L;

        pq1 = Linear_from_ChimInd(ti, tj, 0, tk0);
        pq2 = Linear_from_ChimInd(ti, tj, 1, tk1);

        return;
    }

    coupler = temp_coupler - DW_L*DW_L*DW_M*DW_N;

    # Horizontal Intercell coupler:
    if (coupler < DW_L*DW_M*(DW_N - 1))
    {
        tk1 = coupler % DW_L;
        coupler -= tk1;
        coupler /= DW_L;

        tj = coupler % (DW_N - 1);
        coupler -= tj;
        coupler /= (DW_N - 1);

        ti = coupler;

        pq1 = Linear_from_ChimInd(ti,   tj, 1, tk1);
        pq2 = Linear_from_ChimInd(ti, 1+tj, 1, tk1);

        return;
    }

    coupler = temp_coupler - DW_L*DW_L*DW_M*DW_N - DW_L*DW_M*(DW_N - 1);

    # Vertical Intercell coupler:
    if (coupler < DW_L*(DW_M - 1)*DW_N)
    {
        tk0 = coupler % DW_L;
        coupler -= tk0;
        coupler /= DW_L;

        tj = coupler % DW_N;
        coupler -= tj;
        coupler /= DW_N;

        ti = coupler;

        pq1 = Linear_from_ChimInd(  ti, tj, 0, tk0);
        pq2 = Linear_from_ChimInd(1+ti, tj, 0, tk0);

        return;
    }

    printf "error: coupler %d out of range\n", temp_coupler;
    exit 1;
}
'
