#!/bin/bash

# This shell script validates qbsolv output.  It first parses the QUBO
# file generated by "dw qbsolv" in order to pick up the variables and
# assertions.  Then it parses the qbsolv output and sets the variable
# values, then evaluates each assertion.

usage="dw-val-qbsolv.bash -q <qubo> -o <qbsolv-output>"

columns_per_batch=8

function main
{
    for arg in $*
    do
	[[ -n $grab_qubo ]] && qubo=$arg && unset grab_qubo && continue
	[[ -n $grab_output ]] && qbsolv_output=$arg && unset grab_output && continue
	[[ $arg == -q ]] && grab_qubo=1 && continue
	[[ $arg == -o ]] && grab_output=1 && continue
	weird_line=1
	break
    done

    if [[ -z $qubo || -z $qbsolv_output || -n $weird_line ]]
    then
	printf "\nUSAGE: ${usage}\n\n"
	return 1
    fi

    if [[ ! -e $qubo ]]
    then
	echo "ERROR: cannot read qubo file ${qubo}"
	return 1
    fi

    if [[ ! -e $qbsolv_output ]]
    then
	echo "ERROR: cannot read qbsolv output ${qbsolv_output}"
	return 1
    fi

    read_vars < ${qubo}

    read_asserts < ${qubo}

    read_qbo < ${qbsolv_output}

    set_variables

    evaluate_asserts

    display_variables

    display_assertions
}

function read_vars
{
    while read token1 token2 token3
    do
	if [[ $token1 == VARIABLE* ]]
	then
	    token1=${token1#*\"}
	    token1=${token1%\"*}
	    token2=${token2#*=}
	    var_name[token2]=$token1
	fi
    done
}

function read_asserts
{
    local token1
    local token2
    
    while read token1 token2
    do
	if [[ $token1 == ASSERT* ]]
	then
	    token1=${token1##ASSERT\(}
	    token1=${token1%%\):}
	    token2=${token2##*expr=\"}
	    token2=${token2%%\"}
	    assert[token1]=${token2}
	fi
    done
}

function read_qbo
{
    local line_num

    line_num=1
    while read
    do
	[[ $line_num == 1 ]] && bits=${REPLY%% *}

	[[ $line_num == 2 ]] && solution=${REPLY}

	[[ $line_num == 3 ]] && objective=${REPLY%% *}

	line_num=$((line_num+1))
    done
}

function set_variables
{
    for ((i=0; i<${#var_name[*]}; ++i))
    do
	bit=${solution:${i}:1}
	eval ${var_name[i]}=$bit
    done
}

function evaluate_asserts
{
    local obj_val

    sum=0
    for ((i=0; i<${#assert[*]}; ++i))
    do
	obj_val=$(( ${assert[i]} ))
	assert_val[i]=$(( obj_val == 0 ))
	sum=$((sum + assert_val[i]))
    done
}

function display_variables
{
    echo -e "\n**** SOLUTION ****\n"

    for ((i=0; i<${#var_name[*]}; ++i))
    do
	bit=${solution:${i}:1}
	echo "${var_name[i]} <== $bit"
    done
}

function display_assertions
{
    local displayed
    local num_to_display

    displayed=0
    num_to_display=${#assert[*]}

    (( sum == ${#assert[*]} )) &&
    printf "\nSOLUTION IS VALID\n" ||
    printf "\nSOLUTION IS NOT VALID\n"

    echo -e "\nOBJECTIVE: ${objective}\n\n"

    while (( displayed < num_to_display ))
    do
	local limit

	limit=$((displayed + columns_per_batch))

	(( num_to_display < limit )) && limit=$num_to_display

        for ((i=displayed; i<limit; ++i))
	do
	    printf "V%04d " $i
	done
	printf "\n"

        for ((i=displayed; i<limit; ++i))
	do
	    echo -n "----- "
	done
	printf "\n"

        for ((i=displayed; i<limit; ++i))
	do
	    printf "%5d " ${assert_val[i]}
	done
	printf "\n\n"

	displayed=$limit
    done
}

main $*
